citation = """
@article{gamella2024chamber,
  title={The Causal Chambers: Real Physical Systems as a Testbed for AI Methodology},
  author={Gamella, Juan L. and B\"uhlmann, Peter and Peters, Jonas},
  journal={(work in progress)},
  year={2023}
}
"""

print(
    f"If you use our datasets or models for your work please consider citing:\n{citation}"
)
